"""Type definitions and protocols for ACE-X DevKit."""

from typing import Protocol, Any, Dict, runtime_checkable


@runtime_checkable
class Renderable(Protocol):
    """Protocol for objects that can be rendered to configuration."""
    
    def render(self, model: Dict[str, Any], asset: Any = None) -> Any:
        """Render configuration."""
        ...


@runtime_checkable
class Parseable(Protocol):
    """Protocol for objects that can parse configuration."""
    
    def parse(self, configuration: str) -> Any:
        """Parse configuration."""
        ...


@runtime_checkable
class Transportable(Protocol):
    """Protocol for objects that can transport configuration."""
    
    def connect(self) -> None:
        """Connect to device."""
        ...
    
    def send(self, payload: Any) -> None:
        """Send configuration."""
        ...
    
    def verify(self) -> bool:
        """Verify configuration."""
        ...
    
    def rollback(self) -> None:
        """Rollback configuration."""
        ...




from pydantic import BaseModel, Field
from typing import Optional, Dict, List, Literal, ClassVar, Union, Any
from enum import Enum

from acex_devkit.models.external_value import ExternalValue
from acex_devkit.models.attribute_value import AttributeValue
from acex_devkit.models.container_entry import ContainerEntry
from acex_devkit.models.logging import (
    LoggingConfig,
    Console,
    RemoteServers,
    VtyLine,
    FileLogging,
    LoggingEvents
)
from acex_devkit.models.spanning_tree import SpanningTree
from acex_devkit.models.acl_model import Acl

class MetadataValueType(str, Enum):
    CONCRETE = "concrete"
    EXTERNALVALUE = "externalValue"
    REFERENCE = "reference"

class Metadata(BaseModel):
    type: Optional[str] = "str"
    value_source: MetadataValueType = MetadataValueType.CONCRETE 

class Reference(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("pointer",)
    pointer: str
    metadata: Metadata = Metadata(type="str", value_source="reference")

class ReferenceTo(Reference):
    pointer: str
    metadata: Optional[Dict] = {}

class ReferenceFrom(Reference):
    pointer: str
    metadata: Optional[Dict] = {}

class RenderedReference(BaseModel):
    from_ptr: str
    to_ptr: str

class SystemConfig(BaseModel):
    contact: Optional[AttributeValue[str]] = None
    domain_name: Optional[AttributeValue[str]] = None
    hostname: Optional[AttributeValue[str]] = None
    location: Optional[AttributeValue[str]] = None
    login_banner: Optional[AttributeValue[str]] = None
    motd_banner: Optional[AttributeValue[str]] = None

#class TripleA(BaseModel): ...
# Trying to avoid using "Logging" or "logging" as names for anything due to conflicts with standard lib.

class VtyLines(BaseModel):
    lines: Dict[str, VtyLine] = {}

class LogFiles(BaseModel):
    files: Dict[str, FileLogging] = {}

class LoggingComponents(BaseModel):
    config: LoggingConfig = LoggingConfig()
    console: Optional[Console] = None
    remote_servers: Optional[RemoteServers] = RemoteServers()
    events: Optional[LoggingEvents] = None
    vty: Optional[VtyLines] = VtyLines()
    files: Optional[LogFiles] = LogFiles()

class NtpConfig(BaseModel):
    enabled: AttributeValue[bool] = AttributeValue(value=False)

class NtpServer(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    address: AttributeValue[str]
    port: Optional[AttributeValue[int]] = None
    version: Optional[AttributeValue[int]] = None
    association_typ: Optional[AttributeValue[str]] = None
    prefer: Optional[AttributeValue[bool]] = None
    source_interface: Optional[AttributeValue[str]] = None

class Ntp(BaseModel): 
    config: Optional[NtpConfig] = None
    servers: Optional[Dict[str, NtpServer]] = {}

class SshServer(BaseModel): 
    enable: Optional[AttributeValue[bool]] = None
    protocol_version: Optional[AttributeValue[int]] = AttributeValue(value=2)
    timeout: Optional[AttributeValue[int]] = None
    auth_retries: Optional[AttributeValue[int]] = None
    source_interface: Optional[Reference] = None

class AuthorizedKeyAlgorithms(str, Enum):
    SSH_ED25519 = "ssh-ed25519"
    ECDSA_NISTP256 = "ecdsa-sha2-nistp256"
    ECDSA_NISTP384 = "ecdsa-sha2-nistp384"
    ECDSA_NISTP521 = "ecdsa-sha2-nistp521"
    RSA_SHA2_256 = "rsa-sha2-256"
    RSA_SHA2_512 = "rsa-sha2-512"
    SK_SSH_ED25519 = "sk-ssh-ed25519@openssh.com"
    SK_ECDSA_NISTP256 = "sk-ecdsa-sha2-nistp256@openssh.com"
    SSH_RSA = "ssh-rsa"
    SSH_DSS = "ssh-dss"

class AuthorizedKey(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("public_key",)
    algorithm: Optional[AttributeValue[AuthorizedKeyAlgorithms]] = None
    public_key: Optional[AttributeValue[str]] = None

class Ssh(BaseModel):
    config: Optional[SshServer] = None
    host_keys: Optional[Dict[str, AuthorizedKey]] = {}

class LldpConfigAttributes(BaseModel):
    enabled: Optional[AttributeValue[bool]] = None
    transmit_interval: Optional[AttributeValue[int]] = None
    hold_time: Optional[AttributeValue[int]] = None
    system_name: Optional[AttributeValue[str]] = None  # The system name field shall contain an alpha-numeric string that indicates the system's administratively assigned name.
    suppress_tlv_advertisement: Optional[AttributeValue[List[str]]] = None # Comma-separated list of TLVs to suppress in advertisements. Values depend on device and vendor, but common ones include "system_name", "system_description", "system_capabilities", "management_address", etc.
    interfaces: Optional[Dict[str, Reference]] = {}

class CdpConfigAttributes(BaseModel):
    enabled: Optional[AttributeValue[bool]] = None
    transmit_interval: Optional[AttributeValue[int]] = None
    hold_time: Optional[AttributeValue[int]] = None
    advertise_v2: Optional[AttributeValue[bool]] = None # Whether to advertise CDP version 2
    interfaces: Optional[Dict[str, Reference]] = {}

#class Lldp(BaseModel):
#    config: Optional[LldpConfigAttributes] = LldpConfigAttributes()
#    #interfaces: Optional[Dict[str, LldpInterfaceConfigAttributes]] = {}
#    interfaces: Optional[Dict[str, Reference]] = {}

class Vlan(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("vlan_id",)
    name: AttributeValue[str]
    vlan_id: Optional[AttributeValue[int]] = None
    vlan_name: Optional[AttributeValue[str]] = None
    network_instance: Optional[AttributeValue[str]] = None

class Interface(ContainerEntry, BaseModel):
    "Base class for all interfaces"
    identity_fields: ClassVar[tuple[str, ...]] = ("index", "type")
    index: AttributeValue[int]
    name: AttributeValue[str]

    description: Optional[AttributeValue[str]] = None
    enabled: Optional[AttributeValue[bool]] = None
    ipv4: Optional[AttributeValue[str]] = None
    redirects: Optional[AttributeValue[bool]] = None # Regarding IP redirects. 
    proxy_arp: Optional[AttributeValue[bool]] = None # Cisco true = yes / false = no, Juniper true = unrestricted / false = restricted
    
    type: Literal[
        "ethernetCsmacd",
        "ieee8023adLag",
        "l3ipvlan",
        "softwareLoopback",
        "subinterface",
        "managementInterface"
        ] = "ethernetCsmacd"
    
    model_config = {
        "discriminator": "type"
    }
    

class EthernetCsmacdInterface(Interface):
    "Physical Interface"
    type: Literal["ethernetCsmacd"] = "ethernetCsmacd"

    # Egenskaper för fysiska interface
    stack_index: Optional[AttributeValue[int]] = None
    module_index: Optional[AttributeValue[int]] = None
    subinterfaces: list["SubInterface"] = Field(default_factory=list)
    speed: Optional[AttributeValue[int]] = None
    duplex: Optional[AttributeValue[str]] = None
    switchport: Optional[AttributeValue[bool]] = None
    switchport_mode: Optional[AttributeValue[Literal["access", "trunk"]]] = None
    trunk_allowed_vlans: Optional[AttributeValue[List[int]]] = None
    native_vlan: Optional[AttributeValue[int]] = None
    access_vlan: Optional[AttributeValue[int]] = None
    vlan_id: Optional[AttributeValue[int]] = None
    voice_vlan: Optional[AttributeValue[int]] = None
    mtu: Optional[AttributeValue[int]] = None # No default set as it differs between devices and vendors
    negotiation: Optional[AttributeValue[bool]] = None
    #lldp_enabled: Optional[AttributeValue[bool]] = None
    #cdp_enabled: Optional[AttributeValue[bool]] = None

    # LACP relaterade attribut
    aggregate_id: Optional[AttributeValue[int]] = None
    lacp_enabled: Optional[AttributeValue[bool]] = None
    lacp_mode: Optional[AttributeValue[Literal["active", "passive", "on", "auto"]]] = None
    lacp_port_priority: Optional[AttributeValue[int]] = None
    #lacp_system_id_mac: Optional[AttributeValue[str]] = None
    lacp_interval: Optional[AttributeValue[Literal["fast", "slow"]]] = None

    # Spanning-tree relaterade attribut
    stp_port_priority: Optional[AttributeValue[int]] = None
    stp_cost: Optional[AttributeValue[int]] = None
    stp_edge_port: Optional[AttributeValue[bool]] = None
    stp_bpdu_filter: Optional[AttributeValue[bool]] = None
    stp_bpdu_guard: Optional[AttributeValue[bool]] = None
    stp_loop_guard: Optional[AttributeValue[bool]] = None
    stp_root_guard: Optional[AttributeValue[bool]] = None
    stp_portfast: Optional[AttributeValue[bool]] = None
    stp_link_type: Optional[AttributeValue[Literal["point-to-point", "shared"]]] = None


class Ieee8023adLagInterface(Interface):
    "LAG Interface"
    type: Literal["ieee8023adLag"] = "ieee8023adLag"
    aggregate_id: Optional[AttributeValue[int]] = None
    members: Optional[AttributeValue[List[str]]] = None
    max_ports: Optional[AttributeValue[int]] = None
    switchport: Optional[AttributeValue[bool]] = None
    switchport_mode: Optional[AttributeValue[Literal["access", "trunk"]]] = None
    trunk_allowed_vlans: Optional[AttributeValue[List[int]]] = None
    native_vlan: Optional[AttributeValue[int]] = None
    mtu: Optional[AttributeValue[int]] = None # No default set as it differs between devices and vendors

class L3IpvlanInterface(Interface):
    "SVI Interface"
    type: Literal["l3ipvlan"] = "l3ipvlan"
    vlan_id: Optional[AttributeValue[int]] = None

class SoftwareLoopbackInterface(Interface):
    "Loopback Interface"
    type: Literal["softwareLoopback"] = "softwareLoopback"

    # Loopback har varken vlan, duplex eller speed
    vlan_id: Optional[AttributeValue[int]] = None
    ipv4: Optional[AttributeValue[str]] = None

class SubInterface(Interface):
    "Subinterface"
    type: Literal["subinterface"] = "subinterface"

    vlan_id: Optional[AttributeValue[int]] = None
    ipv4: Optional[AttributeValue[str]] = None

class ManagementInterface(Interface):
    "Management Interface"
    type: Literal["managementInterface"] = "managementInterface"

    # Mgmt har inte vlan
    vlan_id: Optional[AttributeValue[int]] = None

class StaticRouteNextHop(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("next_hop",)
    index: Optional[AttributeValue[int]] = None
    next_hop: AttributeValue[str] # can be an IP address or an interface. Reference will be handled in config component
    metric: Optional[AttributeValue[int]] = None
    static_route: Optional[AttributeValue[str]] = None # Reference to parent static route, used for easier access in config component
    network_instance: Optional[AttributeValue[str]] = None


class StaticRoute(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("prefix",)
    route_name: Optional[AttributeValue[str]] = None
    prefix: AttributeValue[str]
    next_hops: Optional[Dict[str, StaticRouteNextHop]] = {}
    network_instance: Optional[AttributeValue[str]] = None

class Protocols(BaseModel):
    static_routes: Optional[Dict[str, StaticRoute]] = {}
    # OSPF, BGP, etc. can be added here as needed

class RouteTarget(BaseModel):
    value: Optional[AttributeValue[str]] = None

class ImportExportPolicy(BaseModel):
    export_route_target: Optional[List[RouteTarget]] = None
    import_route_target: Optional[List[RouteTarget]] = None

class InterInstancePolicy(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    import_export_policy: ImportExportPolicy

class NetworkInstance(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("name",)
    name: AttributeValue[str]
    description: Optional[AttributeValue[str]] = None
    vlans: Optional[Dict[str, Vlan]] = {}
    interfaces: Optional[Dict[str, Reference]] = {}
    inter_instance_policies: Optional[Dict[str, InterInstancePolicy]] = {}
    protocols: Optional[Protocols] = Protocols()

class LacpConfig(BaseModel):
    system_priority: Optional[AttributeValue[int]] = None
    system_id_mac: Optional[AttributeValue[str]] = None
    load_balance_algorithm: Optional[AttributeValue[list[Literal["src-mac", "dst-mac", "src-dst-mac", "src-ip", "dst-ip", "src-dst-ip", "src-port", "dst-port", "src-dst-port"]]]] = None

class Lacp(BaseModel):
    config: Optional[LacpConfig] = LacpConfig()
    interfaces: Optional[Dict[str, Interface]] = {}

# SNMP
class SnmpAccess(str, Enum):
	READ_ONLY = "READ_ONLY"
	READ_WRITE = "READ_WRITE"


class SnmpSecurityLevel(str, Enum):
	NO_AUTH_NO_PRIV = "NO_AUTH_NO_PRIV"
	AUTH_NO_PRIV = "AUTH_NO_PRIV"
	AUTH_PRIV = "AUTH_PRIV"


class SnmpAuthProtocol(str, Enum):
	MD5 = "MD5"
	SHA1 = "SHA"
	SHA224 = "SHA-224"
	SHA256 = "SHA-256"
	SHA384 = "SHA-384"
	SHA512 = "SHA-512"


class SnmpPrivProtocol(str, Enum):
	DES = "DES"
	TRIPLE_DES = "3DES"
	AES128 = "AES128"
	AES192 = "AES192"
	AES256 = "AES256"


class SnmpConfig(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    enabled: AttributeValue[bool] = AttributeValue(value=False)
    engine_id: Optional[AttributeValue[str]] = None
    location: Optional[AttributeValue[str]] = None
    contact: Optional[AttributeValue[str]] = None


class SnmpCommunity(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("name",)
    name: AttributeValue[str]
    community: Optional[AttributeValue[str]] = None # Community string
    access: Optional[AttributeValue[SnmpAccess]] = AttributeValue(value=SnmpAccess.READ_ONLY)
    view: Optional[AttributeValue[str]] = None
    ipv4_acl: Optional[AttributeValue[str]] = None # Cisco and "liknande" vendors 
    ipv6_acl: Optional[AttributeValue[str]] = None
    source_interface: Optional[Reference] = None
    clients: Optional[AttributeValue[List[str]]] = None # Juniper specific


class SnmpUser(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("username",)
    username: AttributeValue[str]
    security_level: Optional[AttributeValue[SnmpSecurityLevel]] = AttributeValue(value=SnmpSecurityLevel.NO_AUTH_NO_PRIV)
    auth_protocol: Optional[AttributeValue[SnmpAuthProtocol]] = None
    auth_password: Optional[AttributeValue[str]] = None
    priv_protocol: Optional[AttributeValue[SnmpPrivProtocol]] = None
    priv_password: Optional[AttributeValue[str]] = None


class SnmpView(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("name",)
    name: AttributeValue[str]
    oid: AttributeValue[str]
    included: Optional[AttributeValue[bool]] = AttributeValue(value=True)


class SnmpServer(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    name: Optional[AttributeValue[str]] = None
    address: Optional[AttributeValue[str]] = None
    port: Optional[AttributeValue[int]] = AttributeValue(value=162)
    enabled: Optional[AttributeValue[bool]] = AttributeValue(value=True)
    version: Optional[AttributeValue[Literal["v2c", "v3"]]] = None
    community: Optional[AttributeValue[str]] = None
    username: Optional[AttributeValue[str]] = None
    security_level: Optional[AttributeValue[SnmpSecurityLevel]] = None
    source_interface: Optional[Reference] = None
    network_instance: Optional[AttributeValue[str]] = None

# ----------------------------
# Enum-based trap groups
# ----------------------------
# En stor class med Enum	
class TrapEventOptions(str, Enum):
	VRF_UP = "vrf-up"
	VRF_DOWN = "vrf-down"
	VNET_TRUNK_UP = "vnet-trunk-up"
	VNET_TRUNK_DOWN = "vnet-trunk-down"
	ALL = "all"
	RfEvent = 'enabled'
	VlanMembershipEvent = 'enabled'
	ErrdisableEvent = 'enabled'
	CHANGE = "change"
	MOVE = "move"
	THRESHOLD = "threshold"
	AUTHENTICATION = "authentication"
	LINKDOWN = "linkdown"
	LINKUP = "linkup"
	COLDSTART = "coldstart"
	WARMSTART = "warmstart"
	FLOWMONEVENT = "trapflowmonevent"
	ENTITYPERFEVENT = "trapentityperfevent"
	PCALLHOMEEVENT_MESSAGE_SEND_FAIL = "trapcallhomeevent_MESSAGE_SEND_FAIL"
	CALLHOMEEVENT_SERVER_FAIL = "trapcallhomeevent_SERVER_FAIL"	
	TTYEVENT = "trapttyevent"
	EIGRPEVENT = "trapeigrpevent"
	OSPF_STATE_CHANGE = "ospf_state_change"
	OSPF_ERRORS = "ospf_errors"
	OSPF_RETRANSMIT = "ospf_retransmit"
	OSPF_LSA = "ospf_lsa"
	OSPF_CISCO_TRANS_CHANGE = "ospf_cisco_trans_change"
	OSPF_CISCO_SHAMLINK_INTERFACE = "ospf_cisco_shamlink_interface"
	OSPF_CISCO_SHAMLINK_NEIGHBOR = "ospf_cisco_shamlink_neighbor"
	BFD_EVENT = "bfd_event"
	CISCO_SMART_LICENSE_EVENT = "cisco_smart_license_event"
	AUTH_FRAMEWORK_SEC_VIOLATION = "auth_framework_sec_violation"
	REP_EVENT = "rep_event"
	MEMORY_BUFFERPEAK = "memory_bufferpeak"
	CONFIG_COPY = "config_copy"
	CONFIG = "config"
	CONFIG_CTID = "config_ctid"
	ENERGYWISE_EVENT = "energy_wise_event"
	FRU_CTRL_EVENT = "fru_ctrl_event"
	ENTITY_EVENT = "entity_event"
	FLASH_INSERTION = "flash_insertion"
	FLASH_REMOVAL = "flash_removal"
	FLASH_LOWSPACE = "flash_lowspace"
	POWER_ETHERNET_POLICE = "power_ethernet_police"
	POWER_ETHERNET_GROUP_THRESHOLD = "power_ethernet_group_threshold"
	CPU_THRESHOLD = "cpu_threshold"
	SYSLOG = "syslog"
	UDLD_LINK_FAIL_RPT = "udld_link_fail_rpt"
	UDLD_STATUS_CHANGE = "udld_status_change"
	VTP_EVENT = "vtp_event"
	VLAN_CREATE = "vlancreate"
	VLAN_DELETE = "vlandelete"
	PORT_SECURITY = "port_security"
	ENV_MON = "env_mon"
	STACKWISE = "stackwise"
	MVPN = "mvpn"
	PW_VC = "pw_vc"	
	IPSLA = "ipsla"
	DHCP = "dhcp"
	EVENT_MANAGER = "event_manager"
	IKE_POLICY_ADD = "ike_policy_add"
	IKE_POLICY_DELETE = "ike_policy_delete"
	IKE_TUNNEL_START = "ike_tunnel_start"
	IKE_TUNNEL_STOP = "ike_tunnel_stop"
	IPSEC_CRYPTOMAP_ADD = "ipsec_cryptomap_add"
	IPSEC_CRYPTOMAP_DELETE = "ipsec_cryptomap_delete"
	IPSEC_CRYPTOMAP_ATTACH = "ipsec_cryptomap_attach"
	IPSEC_CRYPTOMAP_DETACH = "ipsec_cryptomap_detach"
	IPSEC_TUNNEL_START = "ipsec_tunnel_start"
	IPSEC_TUNNEL_STOP = "ipsec_tunnel_stop"
	IPSEC_TOO_MANY_SAS = "ipsec_too_many_sas"
	OSPFV3_STATE_CHANGE = "ospfv3_state_change"
	OSPFV3_ERRORS = "ospfv3_errors"
	IP_MULTICAST = "ip_multicast"
	MSDP = "msdp"
	PIM_NEIGHBOR_CHANGE = "pim_neighbor_change"
	PIM_RP_MAPPING_CHANGE = "pim_rp_mapping_change"
	INVALID_PIM_MESSAGE = "invalid_pim_message"
	BRIDGE_NEWROOT = "bridge_newroot"
	BRIDGE_TOPOLOGYCHANGE = "bridge_topologychange"
	STPX_INCONSISTENCY = "stpx_inconsistency"
	STPX_ROOT_INCONSISTENCY = "stpx_root_inconsistency"
	STPX_LOOP_INCONSISTENCY = "stpx_loop_inconsistency"
	BGP_CBG2 = "bgp_cbg2"
	HSRP = "hsrp"
	ISIS = "isis"
	CEF_RESOURCE_FAILURE = "cef_resource_failure"
	CEF_PEER_STATE_CHANGE = "cef_peer_state_change"
	CEF_PEER_FIB_STATE_CHANGE = "cef_peer_fib_state_change"
	CEF_INCONSISTENCY = "cef_inconsistency"
	LISP = "lisp"
	NHRP_NHS = "nhrp_nhs"
	NHRP_NHC = "nhrp_nhc"
	NHRP_NHP = "nhrp_nhp"
	NHRP_QUOTA_EXCEEDED = "nhrp_quota_exceeded"
	LOCAL_AUTH = "local_auth"
	ENTITY_DIAG_BOOT_UP_FAIL = "entity_diag_boot_up_fail"
	ENTITY_DIAG_HM_TEST_RECOVER = "entity_diag_hm_test_recover"
	ENTITY_DIAG_HM_THRESH_REACHED = "entity_diag_hm_thresh_reached"
	ENTITY_DIAG_SCHEDULED_TEST_FAIL = "entity_diag_scheduled_test_fail"
	MPLS_RFC_LDP = "mpls_rfc_ldp"
	MPLS_LDP = "mpls_ldp"
	MPLS_RFC_TRAFFIC_ENG = "mpls_rfc_traffic_eng"
	MPLS_TRAFFIC_ENG = "mpls_traffic_eng"
	MPLS_FAST_REROUTE_PROTECTED = "mpls_fast_reroute_protected"
	MPLS_VPN = "mpls_vpn"
	MPLS_RFC_VPN = "mpls_rfc_vpn"
	BULKSTAT_COLLECTION = "bulkstat_collection"
	BULKSTAT_TRANSFER = "bulkstat_transfer"

class TrapEvent(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("event_name",)
    name: Optional[AttributeValue[str]] = None
    event_name: Optional[AttributeValue[TrapEventOptions]] = None

#class SnmpTrap(BaseModel): ...

class Snmp(BaseModel):
    config: Optional[Dict[str, SnmpConfig]] = {}
    communities: Optional[Dict[str, SnmpCommunity]] = {}
    users: Optional[Dict[str, SnmpUser]] = {}
    trap_servers: Optional[Dict[str, SnmpServer]] = {}
    trap_events: Optional[Dict[str, TrapEvent]] = {}
    views: Optional[Dict[str, SnmpView]] = {}

# AAA
class aaaBaseClass(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()  # key is identity (e.g. "default", "console")
    name: Optional[AttributeValue[str]] = None

class aaaTacacsAttributes(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    port: Optional[AttributeValue[int]] = None
    secret_key: Optional[AttributeValue[str]] = None
    secret_key_hashed: Optional[AttributeValue[str]] = None
    address: Optional[AttributeValue[str]] = None
    timeout: Optional[AttributeValue[int]] = None
    source_interface: Optional[Reference] = None
    server_group: Optional[AttributeValue[str]] = None

class aaaRadiusAttributes(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    auth_port: Optional[AttributeValue[int]] = None
    acct_port: Optional[AttributeValue[int]] = None
    secret_key: Optional[AttributeValue[str]] = None
    secret_key_hashed: Optional[AttributeValue[str]] = None
    address: Optional[AttributeValue[str]] = None
    timeout: Optional[AttributeValue[int]] = None
    source_interface: Optional[Reference] = None
    retransmit_attempts: Optional[AttributeValue[int]] = None
    server_group: Optional[AttributeValue[str]] = None

class aaaServerGroupAttributes(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()  # key is the group name
    """
    Define a AAA server group that can contain multiple TACACS+ and/or RADIUS servers.

    Type is used to tell future renderers what kind of server group this is.
    Example:
    type = 'tacacs' or type = 'radius'

    The tacacs and radius attributes expect a reference to the aaaTacacs and aaaRadius models respectively.

    Example in config map:
    enable = True
    type = 'tacacs'
    tacacs = [tacacs_server1, tacacs_server2]
    radius = radius_server1

    Cisco example:
    aaa group server tacacs+ TACACS-GROUP
     server name tacacs_server1
     server name tacacs_server2
    """
    enable: Optional[AttributeValue[bool]] = None
    type: Optional[AttributeValue[Literal['tacacs','radius']]] = None
    tacacs: Optional[Dict[str, aaaTacacsAttributes]] = {}
    radius: Optional[Dict[str, aaaRadiusAttributes]] = {}

# Authentication Models
class aaaAuthenticationMethods(aaaBaseClass):
    """
    Define the authentiation methods used by AAA. If you define a server group using the "aaaServerGroup" model, 
    you can reference it here by its name, but only as a string.

    Example in config map:
    method = ['TACACS_GROUP','LOCAL']

    Cisco example:
    aaa authentication login default group TACACS-GROUP-NEW local
    """
    #method: Optional[List[str]] = None # Ex. ['TACACS_GROUP','LOCAL', 'default', 'enable'] - TACACS_GROUP is reference to server group
    method: Optional[AttributeValue[str]] = None 
    # Method could handle a reference to a server group in the future. For now we only use strings. Important for users to know this.

    # Cisco example:
    # aaa authentication login default group TACACS-GROUP-NEW local
    # aaa authentication enable default group TACACS-GROUP-NEW enable

class authenticationUser(aaaBaseClass):
    identity_fields: ClassVar[tuple[str, ...]] = ("username",)
    username: Optional[AttributeValue[str]] = None
    password: Optional[AttributeValue[str]] = None
    password_hahsed: Optional[AttributeValue[str]] = None
    ssh_key: Optional[AttributeValue[str]] = None
    role: Optional[AttributeValue[str]] = None

class aaaAuthenticationUsers(aaaBaseClass):
    username: Optional[Dict[str, authenticationUser]] = {}

class adminUser(aaaBaseClass):
    admin_password: Optional[AttributeValue[str]] = None
    admin_password_hashed: Optional[AttributeValue[str]] = None

class aaaAuthenticationAdminUsers(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    config: Optional[Dict[str, adminUser]] = {}

class aaaAuthentication(BaseModel):
    config: Optional[Dict[str, aaaAuthenticationMethods]] = {}
    admin_user: Optional[Dict[str, aaaAuthenticationAdminUsers]] = {}
    users: Optional[Dict[str, aaaAuthenticationUsers]] = {}

# Authorization Models
class aaaAuthorizationMethods(aaaBaseClass):
    """
    Define the authorization methods used by AAA. If you define a server group using the "aaaServerGroup" model, 
    you can reference it here by its name, but only as a string.

    Example in config map:
    method = ['TACACS_GROUP','LOCAL']

    Cisco example:
    aaa authorization login default group TACACS-GROUP-NEW local
    """
    #method: Optional[List[str]] = None # Ex. ['TACACS_GROUP','LOCAL']
    method: Optional[AttributeValue[str]] = None

class aaaAuthorizationEvents(aaaBaseClass):
    """
    Define authorization events. 

    Cisco example:
    aaa authorization config-commands
    aaa authorization console
    """
    #events: Optional[List[str]] = Field(default_factory=list) # Ex. ['config-commands','console']
    #event: Optional[AttributeValue[List[str]]] = None
    event: Optional[AttributeValue[str]] = None

class aaaAuthorization(BaseModel):
    config: Optional[Dict[str, aaaAuthorizationMethods]] = {}
    events: Optional[Dict[str, aaaAuthorizationEvents]] = {}

# Accounting Models
class aaaAccountingMethods(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    """
    Define the accounting methods used by AAA. If you define a server group using the "aaaServerGroup" model,
    you can reference it here by its name, but only as a string.

    Example in config map:
    method = ['TACACS_GROUP','LOCAL']

    Cisco example:
    aaa accounting login default group TACACS-GROUP-NEW local
    """
    #method: Optional[List[str]] = None # Ex. ['TACACS_GROUP','LOCAL']
    method: Optional[AttributeValue[str]] = None

class aaaAccountingEvents(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    """
    Define accounting events.

    Cisco example:
    aaa accounting send stop-record authentication failure
    """
    #events: Optional[List[str]] = Field(default_factory=list) # Ex. ['send','stop-record','authentication', 'failure']
    event: Optional[AttributeValue[str]] = None

class aaaAccounting(BaseModel):
    config: Optional[Dict[str, aaaAccountingMethods]] = {}
    events: Optional[Dict[str, aaaAccountingEvents]] = {}

class aaaGlobalAttributes(BaseModel):
    enabled: Optional[AttributeValue[bool]] = False # default False

class TripleA(BaseModel):
    config: aaaGlobalAttributes = aaaGlobalAttributes()
    server_groups: Optional[Dict[str, aaaServerGroupAttributes]] = {}
    authentication: aaaAuthentication = aaaAuthentication()
    authorization: aaaAuthorization = aaaAuthorization()
    accounting: aaaAccounting = aaaAccounting()

class VTPAttributes(BaseModel):
    domain_name: Optional[AttributeValue[str]] = None
    mode: Optional[AttributeValue[Literal["server", "client", "transparent", 'off']]] = None
    primary_server: Optional[AttributeValue[bool]] = False # Cisco proprietary
    version: Optional[AttributeValue[Literal[1, 2, 3]]] = None
    password: Optional[AttributeValue[str]] = None
    password_hashed: Optional[AttributeValue[str]] = None

class VTP(BaseModel):
    config: VTPAttributes = VTPAttributes()

class DHCPSnoopingAttributes(BaseModel):
    enabled: Optional[AttributeValue[bool]] = None
    vlans: Optional[Dict[str, Reference]] = {} # VLANs where DHCP snooping is enabled, key is VLAN ID, value is reference to VLAN
    trust_interfaces: Optional[Dict[str, Reference]] = {} # Interfaces that are trusted for DHCP snooping, key is interface name, value is reference to interface
    option82: Optional[AttributeValue[bool]] = None # Whether DHCP snooping option 82 is enabled

class DhcpRelayServerAttributes(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    address: Optional[AttributeValue[str]] = None
    network_instance: Optional[AttributeValue[str]] = None
    interfaces: Optional[Dict[str, Reference]] = {}

class DhcpRelay(BaseModel):
    relay_servers: Optional[Dict[str, DhcpRelayServerAttributes]] = {}

class Dhcp(BaseModel): 
    snooping: Optional[DHCPSnoopingAttributes] = DHCPSnoopingAttributes()
    relay: Optional[DhcpRelay] = DhcpRelay()

class Services(BaseModel):
    name: Optional[AttributeValue[str]] = None
    http: Optional[AttributeValue[bool]] = None # for webgui access
    https: Optional[AttributeValue[bool]] = None # for webgui access

class NetflowFormat(str, Enum):
    IPFIX = "IPFIX"
    NETFLOW_V9 = "NetFlow v9"
    NETFLOW_V5 = "NetFlow v5"

class NetflowRecordIpv4Match(BaseModel):
    # Leaf-level ipv4 matches (Cisco style "match ipv4 <field>"). True = match this field, False = No. None = Ignore the field
    netflow_record: Optional[AttributeValue[str]] = None # Reference to parent record, used for easier access in config component
    dscp: Optional[AttributeValue[bool]] = None
    fragmentation: Optional[AttributeValue[bool]] = None
    header_length: Optional[AttributeValue[bool]] = None
    id: Optional[AttributeValue[bool]] = None
    length: Optional[AttributeValue[bool]] = None
    option: Optional[AttributeValue[bool]] = None
    precedence: Optional[AttributeValue[bool]] = None
    protocol: Optional[AttributeValue[bool]] = None
    section: Optional[AttributeValue[str]] = None
    tos: Optional[AttributeValue[bool]] = None
    total_length: Optional[AttributeValue[bool]] = None
    ttl: Optional[AttributeValue[bool]] = None
    version: Optional[AttributeValue[bool]] = None

class NetflowRecordAttributes(BaseModel): # Cisco flow record
    match_ipv4: Optional[NetflowRecordIpv4Match] = None
    #match_ipv4: Optional[Dict[str, NetflowRecordIpv4Match]] = {}
    application_name: Optional[AttributeValue[bool]] = None

    # Escape hatch for vendor-specific match knobs not yet modeled
    match_vendor_specific: Optional[AttributeValue[Dict[str, Any]]] = None # keeping a flexible option if needed. Not advertised to users.

    collect_timestamp_absolute_first: Optional[AttributeValue[bool]] = None
    collect_timestamp_absolute_last: Optional[AttributeValue[bool]] = None

class NetflowCollectorAttributes(BaseModel): # Cisco flow monitor
    netflow_record: Optional[AttributeValue[str]] = None # Reference to record, used for easier access in config component
    netflow_exporter: Optional[AttributeValue[str]] = None # Reference to exporter, used for easier access in config component
    cache_inactive: Optional[AttributeValue[int]] = None
    cache_active: Optional[AttributeValue[int]] = None
    interfaces: Optional[Dict[str, Reference]] = {} # allow for disabling netflow on specific interfaces

class NetflowExporterOptions(BaseModel):
    # Mostly timeouts atm
    interface_table_timeout: Optional[AttributeValue[int]] = None
    vrf_table_timeout: Optional[AttributeValue[int]] = None
    sampler_table: Optional[AttributeValue[bool]] = None
    application_table_timeout: Optional[AttributeValue[int]] = None
    application_attributes_timeout: Optional[AttributeValue[int]] = None
    netflow_exporter: Optional[AttributeValue[str]] = None # Reference to parent exporter, used for easier access in config component

class NetflowExporterAttributes(ContainerEntry, BaseModel): # Cisco flow exporter
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    address: Optional[AttributeValue[str]] = None
    port: Optional[AttributeValue[int]] = None
    netflow_format: Optional[AttributeValue[str]] = None
    source_interface: Optional[Reference] = None
    network_instance: Optional[AttributeValue[str]] = None
    options: Optional[NetflowExporterOptions] = None

class NetflowGlobalConfigAttributes(BaseModel):
    enabled: Optional[AttributeValue[bool]] = None
    version: Optional[AttributeValue[int]] = None

class Netflow(BaseModel):
    config: Optional[NetflowGlobalConfigAttributes] = None
    records: Optional[Dict[str, NetflowRecordAttributes]] = {}
    exporters: Optional[Dict[str, NetflowExporterAttributes]] = {}
    collectors: Optional[Dict[str, NetflowCollectorAttributes]] = {}

class SflowMonitoringAttributes(BaseModel): ...

class SflowForwardingAttributes(BaseModel): ... # collector?

class SflowCollectorAttributes(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ("address",)
    address: Optional[AttributeValue[str]] = None # Destination IP address of the sFlow collector
    port: Optional[AttributeValue[int]] = None # UDP
    source_address: Optional[AttributeValue[str]] = None 
    network_instance: Optional[AttributeValue[str]] = None # VRF
    interfaces: Optional[Dict[str, Reference]] = {} # allow for disabling sflow on specific interfaces

class SfloGlobalConfigAttributes(BaseModel):
    enabled: Optional[AttributeValue[bool]] = None
    #version: Optional[AttributeValue[int]] = None
    dscp: Optional[AttributeValue[int]] = None # range: 0..63 
    sample_size: Optional[AttributeValue[int]] = None # Sets the maximum number of bytes to be copied from a sampled packet (content within one specific sample of a packet).
    polling_interval: Optional[AttributeValue[int]] = None # seconds
    ingress_sampling_rate: Optional[AttributeValue[int]] = None # sampling rate is 1/N packets. An implementation may implement the sampling rate as a statistical average, rather than a strict periodic sampling.
    egress_sampling_rate: Optional[AttributeValue[int]] = None  # sampling rate is 1/N packets. An implementation may implement the sampling rate as a statistical average, rather than a strict periodic sampling.

class Sflow(BaseModel):
    #enabled: Optional[AttributeValue[bool]] = None
    #version: Optional[AttributeValue[int]] = None
    #dscp: Optional[AttributeValue[int]] = None # range: 0..63 
    #sample_size: Optional[AttributeValue[int]] = None
    #polling_interval: Optional[AttributeValue[int]] = None
    #ingress_sampling_rate: Optional[AttributeValue[int]] = None
    #egress_sampling_rate: Optional[AttributeValue[int]] = None
    collector: Optional[Dict[str, SflowCollectorAttributes]] = {}
    # Similar structure to Netflow can be implemented here for sFlow-specific attributes

class Sampling(BaseModel):
    netflow: Optional[Netflow] = Netflow()
    #sflow: Optional[Sflow] = Sflow() # Sflow can be added in the future, similar structure to Netflow

class System(BaseModel):
    config: SystemConfig = SystemConfig()
    aaa: Optional[TripleA] = TripleA()
    logging: Optional[LoggingComponents] = LoggingComponents() # Trying to avoid using "Logging" or "logging" as names for anything due to conflicts with standard lib.
    ntp: Optional[Ntp] = Ntp()
    ssh: Optional[Ssh] = Ssh()
    snmp: Optional[Snmp] = Snmp()
    vtp: Optional[VTP] = VTP()
    dhcp: Optional[Dhcp] = Dhcp()
    services: Optional[Services] = Services()

# For different types of interfaces that are fine for response model:
InterfaceType = Union[
    EthernetCsmacdInterface,
    Ieee8023adLagInterface,
    L3IpvlanInterface,
    SoftwareLoopbackInterface,
    SubInterface,
    ManagementInterface,
]

class ComposedConfiguration(BaseModel):
    system: Optional[System] = System()
    acl: Optional[Acl] = Acl()
    lldp: Optional[LldpConfigAttributes] = LldpConfigAttributes()
    cdp: Optional[CdpConfigAttributes] = CdpConfigAttributes()
    lacp: Optional[Lacp] = Lacp()
    interfaces: Dict[str, InterfaceType] = {}
    network_instances: Dict[str, NetworkInstance] = {"global": NetworkInstance(name="global")}
    stp: Optional[SpanningTree] = SpanningTree()
    sampling: Optional[Sampling] = Sampling()

"""
GUIDELINES FOR COMPOSED CONFIGURATION:

1. All values must always be typed as AttributeValue.
2. Containers must always be defined as hierarchical pydantic types, no dicts as placeholders.
3. Component collections must use a typed container class (e.g. RemoteServers, VtyLines) with an inner
   Dict[str, BaseModel] field — the key identifies the component. Raw Dict fields are not allowed at
   the container level.
4. Default values for Optional containers are always an empty instance of the container class, e.g.
   Optional[RemoteServers] = RemoteServers(). Never default to None or a raw empty dict for collections.
"""


from pydantic import BaseModel, ConfigDict, field_validator, model_validator, field_serializer
from typing import Union, TypeVar, Generic, Optional, Dict, Any, get_args, get_origin
from acex_devkit.models import ExternalValue
import ipaddress

T = TypeVar('T')

class AttributeValue(BaseModel, Generic[T]):
    """
    A generic wrapper for values that may be concrete or external.
    """
    model_config = ConfigDict(arbitrary_types_allowed=True)

    value: Union[T, ExternalValue]
    metadata: Optional[Dict[str, Any]] = None

    # ---------------------------------------------------------
    # 1) PRE-PROCESSOR: Tillåt råa värden, ExternalValue direkt,
    #    eller dict → AttributeValue
    # ---------------------------------------------------------
    @model_validator(mode="before")
    @classmethod
    def preprocess_raw(cls, data):
        """
        Gör varje typ av input till en fullvärdig {"value": ...} dict
        så att Pydantic kan fortsätta normalt.
        """
        # 1) Rått värde (str, int, ExternalValue etc)
        if not isinstance(data, dict):
            return {"value": data}

        # 2) Dict som representerar ExternalValue
        if "value" not in data and "ref" in data:
            return {"value": ExternalValue(**data)}

        # 3) Dict som redan har value → låt vara
        return data

    # ---------------------------------------------------------
    # 2) VALIDATOR för value
    # ---------------------------------------------------------
    @field_validator("value", mode="before")
    @classmethod
    def normalize_value(cls, v):
        if isinstance(v, cls):
            return v.value
        if isinstance(v, ExternalValue):
            return v
        if isinstance(v, dict) and "ref" in v:
            return ExternalValue(**v)
        return v

    # ---------------------------------------------------------
    # 3) METADATA-generator (after)
    #    Bevarar redan satt metadata!
    # ---------------------------------------------------------
    @model_validator(mode="after")
    def set_automatic_metadata(self):
        """
        Lägg till metadata utan att ta bort användarens egna.
        """
        self.metadata = dict(self.metadata or {})  # KOPIA – ändra inte originalet

        if isinstance(self.value, ExternalValue):
            self.metadata.setdefault("value_type", "external")
            self.metadata.setdefault("attr_ptr", self.value.attr_ptr)
            self.metadata.setdefault("plugin", self.value.plugin)
            self.metadata.setdefault(
                "ev_type",
                self.value.ev_type.value
                if hasattr(self.value.ev_type, "value")
                else self.value.ev_type,
            )
            self.metadata.setdefault("query", self.value.query)
            self.metadata.setdefault("resolved", self.value.resolved)

            if self.value.kind:
                self.metadata.setdefault("kind", self.value.kind)

            if self.value.resolved and self.value.resolved_at:
                self.metadata.setdefault(
                    "resolved_at",
                    self.value.resolved_at.isoformat()
                    if hasattr(self.value.resolved_at, "isoformat")
                    else str(self.value.resolved_at),
                )

        else:
            self.metadata.setdefault("value_type", "concrete")
            self.metadata.setdefault("type", type(self.value).__name__)

        return self

    # ---------------------------------------------------------
    # 4) SERIALIZER – external value → returnera dess "value"
    # ---------------------------------------------------------
    @field_serializer("value")
    def serialize_value(self, value):
        if isinstance(value, ExternalValue):
            return value.value
        # Ensure ipaddress objects serialize cleanly to JSON
        if isinstance(
            value,
            (
                ipaddress.IPv4Address,
                ipaddress.IPv6Address,
                ipaddress.IPv4Network,
                ipaddress.IPv6Network,
                ipaddress.IPv4Interface,
                ipaddress.IPv6Interface,
            ),
        ):
            return str(value)
        return value

    def __eq__(self, other):
        if not isinstance(other, AttributeValue):
            return NotImplemented
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

    # Convenience helpers
    def is_external(self) -> bool:
        return isinstance(self.value, ExternalValue)

    def get_value(self) -> T:
        return self.value.value if isinstance(self.value, ExternalValue) else self.value




from abc import ABC, abstractmethod
from typing import Any, Dict

from acex_devkit.models.node_response import NodeListItem
from acex_devkit.models.management_connection import ManagementConnection

class ParserBase(ABC):
    @abstractmethod
    def parse(self, model: Dict[str, Any]) -> Any:
        """Parsar running-config"""

class RendererBase(ABC):
    @abstractmethod
    def render(self, model: Dict[str, Any]) -> Any:
        """Tar en device‑agnostisk konfigurationsmodell och returnerar
        en transport‑specifik representation (t.ex. string, XML‑tree…)."""

class TransportBase(ABC):
    @abstractmethod
    def get_config(self, node: NodeListItem, connection: ManagementConnection, **kwargs) -> str:
        """Fetch the full running configuration from a device."""
        pass

    @abstractmethod
    def send_config(self, node: NodeListItem, connection: ManagementConnection, commands: list[str], **kwargs) -> str:
        """Apply configuration commands to a device."""
        pass

    def execute(self, node: NodeListItem, connection: ManagementConnection, commands: list[str], **kwargs) -> list[str]:
        """Run arbitrary commands. Opt-in per driver."""
        raise NotImplementedError(f"{self.__class__.__name__} does not implement execute()")

class NetworkElementDriver:
    """Kombinerar renderer + transport – exponeras som en plugin."""
    renderer_class = None
    transport_class = None
    parser_class = None

    def __init__(self):
        if self.renderer_class is None or self.transport_class is None or self.parser_class is None:
            raise NotImplementedError("renderer_class and transport_class must be set in subclass")
        self.renderer = self.renderer_class()
        self.transport = self.transport_class()
        self.parser = self.parser_class()

    @abstractmethod
    def render(self, configuration: ComposedConfiguration, asset: "Asset") -> Any:
        """Render configuration from composedconfig and asset."""
        pass

    @abstractmethod
    def parse(self, configuration: str) -> ComposedConfiguration:
        """
        Parse observed configuration into a composedconfiguration object.
        """

    



from typing import ClassVar


class ContainerEntry:
    """
    Mixin for Pydantic models that are used as values in Dict[str, X] container fields.

    Every subclass MUST declare `identity_fields` as a ClassVar[tuple[str, ...]]:
      - Non-empty tuple: fields whose values uniquely identify this object.
        The differ will match objects across desired/observed by these field values,
        ignoring the dict key entirely.
      - Empty tuple `()`: the dict key itself is the identity (key-based matching).
        Use this for singleton-like entries (e.g., "global", "default") where no
        field inside the object acts as a natural identifier.

    Example:
        class NtpServer(ContainerEntry, BaseModel):
            identity_fields: ClassVar[tuple[str, ...]] = ("address",)
            address: AttributeValue[str]
            ...

        class SpanningTreeGlobalConfig(ContainerEntry, BaseModel):
            identity_fields: ClassVar[tuple[str, ...]] = ()  # key is identity
            mode: Optional[AttributeValue[str]] = None
            ...
    """

    identity_fields: ClassVar[tuple[str, ...]]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        has_identity_fields = any(
            "identity_fields" in c.__dict__
            for c in cls.__mro__
            if c is not ContainerEntry
        )
        if not has_identity_fields:
            raise TypeError(
                f"{cls.__name__} must declare 'identity_fields' as a ClassVar. "
                f"Use identity_fields = ('field_name',) to match by field value, "
                f"or identity_fields = () to match by dict key."
            )



"""Base classes for ACE-X network element drivers."""

from acex_devkit.drivers.base import (
    NetworkElementDriver,
    TransportBase,
    RendererBase,
    ParserBase,
)

__all__ = [
    "NetworkElementDriver",
    "TransportBase",
    "RendererBase",
    "ParserBase",
]



from pydantic import BaseModel
from acex_devkit.models.attribute_value import AttributeValue
from acex_devkit.models.container_entry import ContainerEntry
from enum import Enum
from typing import ClassVar, Optional, Dict

class SpanningTreeGlobalAttributes(ContainerEntry, BaseModel):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    mode: Optional[AttributeValue[str]] = None # Needs to be defined by user. Default for Cisco is RAPID-PVST and for Juniper it's just RSTP
    bpdu_filter: Optional[AttributeValue[bool]] = None # Disabled by default
    bpdu_guard: Optional[AttributeValue[bool]] = None # Disabled by default
    loop_guard: Optional[AttributeValue[bool]] = None # Disabled by default
    portfast: Optional[AttributeValue[bool]] = None # Disabled by default. Global setting for access ports.
    bridge_assurance: Optional[AttributeValue[bool]] = None # Disabled by default. Only supported by MST and PVRST+
    #interfaces: Optional[Dict[str, Reference]] = None

class SpanningTreeModeConfig(BaseModel):
    hello_time: Optional[AttributeValue[int]] = None
    max_age: Optional[AttributeValue[int]] = None
    forward_delay: Optional[AttributeValue[int]] = None
    bridge_priority: Optional[AttributeValue[int]] = None
    hold_count: Optional[AttributeValue[int]] = None # Range 1..10

## RSTP
class RstpAttributes(ContainerEntry, SpanningTreeModeConfig):
    identity_fields: ClassVar[tuple[str, ...]] = ()

class RSTPConfig(BaseModel):
    #config: RstpAttributes = RstpAttributes()
    config: Optional[Dict[str, RstpAttributes]] = None

### MSTP
class MstpInstanceAttributes(ContainerEntry, SpanningTreeModeConfig):
    identity_fields: ClassVar[tuple[str, ...]] = ("instance_id",)
    instance_id: AttributeValue[int] # range: 1..4094
    name: Optional[AttributeValue[str]] = None
    vlan: Optional[AttributeValue[list[int]]] = None # List of VLANs mapped to the MST instance

class MstpAttributes(ContainerEntry, SpanningTreeModeConfig):
    identity_fields: ClassVar[tuple[str, ...]] = ()
    revision: Optional[AttributeValue[int]] = None
    max_hop: Optional[AttributeValue[int]] = None # Range 1..255

class MSTPConfig(BaseModel):
    #config: MstpAttributes = MstpAttributes()
    config: Optional[Dict[str, MstpAttributes]] = None
    mst_instances: Optional[Dict[str, MstpInstanceAttributes]] = None

### Rapid PVST
class RapidPVSTAttributes(ContainerEntry, SpanningTreeModeConfig):
    identity_fields: ClassVar[tuple[str, ...]] = ("vlan",)
    """
    Docstring for RapidPVSTAttributes
    vlan can be a string or list. Depending on how NED is built it will check wether it's a single VLAN or multiple VLANs and then format the data to the correct format
    for the command of the specific vendor.
    For example for Cisco:
    * Single VLAN
        spanning-tree vlan 10 priority 8192
    * Multiple VLANs
        spanning-tree vlan 10-30 priority 8192
    """
    #vlan_id: Optional[AttributeValue[int]] = None  # Single VLAN ID or list of VLANs using Rapid PVST+
    vlan: Optional[AttributeValue[int | list[int]]] = None  # Single VLAN ID or list of VLANs using Rapid PVST+

class RapidPVSTConfig(BaseModel):
    vlan: Optional[Dict[str, RapidPVSTAttributes]] = None

class SpanningTree(BaseModel):
    config: Optional[Dict[str, SpanningTreeGlobalAttributes]] = None#SpanningTreeGlobalAttributes()
    rstp: Optional[RSTPConfig] = RSTPConfig()
    mstp: Optional[MSTPConfig] = MSTPConfig()
    rapidpvst: Optional[RapidPVSTConfig] = RapidPVSTConfig()


"""Driver base classes for ACE-X network element drivers."""

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Dict

from acex_devkit.models.node_response import NodeListItem
from acex_devkit.models.management_connection import ManagementConnection


class ParserBase(ABC):
    """Base class for configuration parsers."""
    
    @abstractmethod
    def parse(self, configuration: str) -> Any:
        """Parse device configuration into a structured model.
        
        Args:
            configuration: Raw configuration string from device
            
        Returns:
            Parsed configuration model
        """
        pass


class RendererBase(ABC):
    """Base class for configuration renderers."""
    
    @abstractmethod
    def render(self, model: Dict[str, Any], asset: Any = None) -> Any:
        """Render configuration model to device-specific format.
        
        Args:
            model: Device-agnostic configuration model
            asset: Optional asset context
            
        Returns:
            Device-specific configuration (e.g., string, XML tree)
        """
        pass


class TransportBase(ABC):
    """Base class for device transport/communication.

    Each method is self-contained, but transports may also expose
    `session()` as a context manager so a caller can run multiple
    operations against the same physical connection.

    Args:
        node: The node instance (identity, hostname, vendor, os, ned_id)
        connection: Management connection (target_ip, connection_type)
        **kwargs: Future use (credentials, options, etc.)
    """

    @contextmanager
    def session(self, connection: ManagementConnection, **kwargs):
        """Open a reusable session for multiple ops on one device.

        Default is a no-op: methods continue to open/close per call.
        Drivers that benefit from connection reuse override this to
        hold a single connection open for the duration of the block.
        """
        yield self

    @abstractmethod
    def get_config(self, node: NodeListItem, connection: ManagementConnection, **kwargs) -> str:
        """Fetch the full running configuration from a device."""
        pass

    @abstractmethod
    def send_config(self, node: NodeListItem, connection: ManagementConnection, commands: list[str], **kwargs) -> str:
        """Apply configuration commands to a device."""
        pass

    def execute(self, node: NodeListItem, connection: ManagementConnection, commands: list[str], **kwargs) -> list[str]:
        """Run arbitrary commands and return output per command. Opt-in per driver."""
        raise NotImplementedError(f"{self.__class__.__name__} does not implement execute()")

    def get_lldp_neighbors(self, node: NodeListItem, connection: ManagementConnection, **kwargs) -> list[dict]:
        """Fetch LLDP/CDP neighbor table from a device.

        Returns list of dicts with keys:
            local_interface, remote_device, remote_interface, discovery_protocol
        """
        raise NotImplementedError(f"{self.__class__.__name__} does not implement get_lldp_neighbors()")


class NetworkElementDriver:
    """Base class for network element drivers.

    Combines renderer, transport, and parser to provide complete
    configuration management for network devices.

    Attributes:
        renderer_class: Renderer class to use (must be set in subclass)
        transport_class: Transport class to use (must be set in subclass)
        parser_class: Parser class to use (must be set in subclass)
        normalizer_class: Optional normalizer class for stripping
            non-intent config and masking secrets
    """

    renderer_class = None
    transport_class = None
    parser_class = None
    normalizer_class = None

    def __init__(self):
        """Initialize driver with renderer, transport, and parser instances."""
        if self.renderer_class is None or self.transport_class is None or self.parser_class is None:
            raise NotImplementedError(
                "renderer_class, transport_class, and parser_class must be set in subclass"
            )
        self.renderer = self.renderer_class()
        self.transport = self.transport_class()
        self.parser = self.parser_class()
        self.normalizer = self.normalizer_class() if self.normalizer_class else None

    @abstractmethod
    def render(self, logical_node: "LogicalNode", asset: Any = None) -> Any:
        """Render logical node to device configuration.
        
        Args:
            logical_node: Logical node containing configuration
            asset: Optional asset context
            
        Returns:
            Rendered configuration
        """
        return self.renderer.render(logical_node.model_dump(), asset)

    @abstractmethod
    def parse(self, configuration: str) -> Any:
        """Parse device configuration.

        Args:
            configuration: Raw device configuration

        Returns:
            Parsed configuration model
        """
        return self.parser.parse(configuration)

    def normalize(self, raw: str) -> str:
        """Strip non-intent data (timestamps, auto-generated certs, etc.).

        Returns the cleaned config string. If no normalizer is configured
        the input is returned unchanged.
        """
        if self.normalizer is None:
            return raw
        return self.normalizer.normalize(raw).config

    def mask(self, raw: str) -> str:
        """Replace secrets with <REDACTED>.

        Returns the masked config string. If no normalizer is configured
        the input is returned unchanged.
        """
        if self.normalizer is None:
            return raw
        return self.normalizer.mask(raw).config



"""
Component-based diff model for comparing ComposedConfiguration objects.

This module provides a structured way to identify which ConfigComponents have been
added, removed, or changed when comparing two configurations. This enables drivers
to render device-specific command patches based on the component changes.
"""

from pydantic import BaseModel, Field, field_serializer, computed_field
from typing import Any, Dict, List, Optional, Type
from enum import Enum

from acex_devkit.models.attribute_value import AttributeValue


class ComponentDiffOp(str, Enum):
    """Operation type for a component change"""
    ADD = "add"
    REMOVE = "remove"
    CHANGE = "change"


class AttributeChange(BaseModel):
    """Represents a change to a single attribute within a component"""
    model_config = {"arbitrary_types_allowed": True}

    attribute_name: str
    before: Optional[Any] = None
    after: Optional[Any] = None


class ComponentChange(BaseModel):
    """
    Represents a change to a ConfigComponent.
    
    This is the core structure that drivers will use to render patches.
    It contains:
    - The component path (e.g., ['interfaces', 'GigabitEthernet0-0-1'])
    - The operation type (add/remove/change)
    - The component class name (e.g., 'FrontpanelPort', 'Loopback')
    - The full before/after component data (serialized dicts)
    - The actual ConfigComponent objects (for driver access)
    - Individual attribute changes (for CHANGE operations)
    """
    op: ComponentDiffOp
    path: List[str]  # e.g., ['interfaces', 'GigabitEthernet0-0-1']
    component_type: Type[Any]  # e.g., 'FrontpanelPort', 'Loopback', 'Vlan'
    component_name: str  # e.g., 'GigabitEthernet0-0-1', 'Lo0', '100'
    
    # Actual ConfigComponent objects
    before: Optional[Any] = None
    after: Optional[Any] = None

    # Raw dict representations (for formatters and generic processing)
    before_dict: Optional[Dict[str, Any]] = None
    after_dict: Optional[Dict[str, Any]] = None
    
    # For CHANGE operations: list of changed attributes
    changed_attributes: List[AttributeChange] = Field(default_factory=list)
    
    model_config = {"arbitrary_types_allowed": True}

    @field_serializer("component_type")
    def serialize_component_type(self, v: Type[Any]) -> str:
        return v.__name__

    @property
    def component_type_name(self) -> str:
        """Return the class name of the component type, e.g., 'FrontpanelPort'"""
        return self.component_type.__name__

    def get_path_str(self, separator: str = "/") -> str:
        """Return the component path as a string, e.g., 'interfaces/GigabitEthernet0-0-1'"""
        return separator.join(self.path) if self.path else ""


class Diff(BaseModel):
    """
    Result of comparing two ComposedConfiguration objects.
    
    This provides a component-centric view of what changed, suitable for
    rendering device-specific patches.
    """
    added: List[ComponentChange] = Field(default_factory=list)
    removed: List[ComponentChange] = Field(default_factory=list)
    changed: List[ComponentChange] = Field(default_factory=list)

    # Total number of config points in desired/observed configs — populated by ConfigDiffer
    total_desired: int = 0
    total_observed: int = 0

    @computed_field
    @property
    def compliant_count(self) -> int:
        """
        Number of desired config points that are fully compliant (present and
        identical on the device). Equals total desired minus points that are
        missing (added) or incorrect (changed).
        """
        return max(0, self.total_desired - len(self.added) - len(self.changed))

    @computed_field
    @property
    def compliance_percentage(self) -> float:
        """
        Percentage of desired config points that are compliant.
        Returns 100.0 when there are no desired config points.
        """
        if self.total_desired == 0:
            return 100.0
        return round(self.compliant_count / self.total_desired * 100, 2)

    def is_empty(self) -> bool:
        """Check if there are no changes"""
        return not (self.added or self.removed or self.changed)

    def summary(self) -> Dict[str, Any]:
        """Get a summary of changes and compliance metadata"""
        return {
            "added": len(self.added),
            "removed": len(self.removed),
            "changed": len(self.changed),
            "total_changes": len(self.added) + len(self.removed) + len(self.changed),
            "total_desired": self.total_desired,
            "total_observed": self.total_observed,
            "compliant_count": self.compliant_count,
            "compliance_percentage": self.compliance_percentage,
        }

    def get_all_changes(self) -> List[ComponentChange]:
        """Get all changes in a single list"""
        return self.added + self.removed + self.changed

    def get_changes_by_type(self, component_type: str) -> List[ComponentChange]:
        """Get all changes for a specific component type (e.g., 'Loopback')"""
        return [
            change for change in self.get_all_changes()
            if change.component_type.__name__ == component_type
        ]

    def get_changes_by_path_prefix(self, path_prefix: List[str]) -> List[ComponentChange]:
        """
        Get all changes under a specific path prefix.
        
        Example:
            get_changes_by_path_prefix(['interfaces']) -> all interface changes
            get_changes_by_path_prefix(['network_instances', 'default']) -> changes in default VRF
        """
        return [
            change for change in self.get_all_changes()
            if change.path[:len(path_prefix)] == path_prefix
        ]



from pydantic import BaseModel, Field
from typing import Annotated, Optional, Dict, Any, List, Literal, Union
from enum import Enum
from datetime import datetime
from acex_devkit.models.composed_configuration import ComposedConfiguration


class LogicalNodeResponse(BaseModel):
    id: Optional[int] = None
    hostname: Optional[str] = None
    role: Optional[str] = None
    site: Optional[str] = None
    sequence: Optional[int] = None
    configuration: Optional[ComposedConfiguration] = None
    meta_data: Optional[Dict[str, Any]] = None


class Asset(BaseModel):
    id: Optional[int] = None
    type: Literal["asset"] = "asset"
    vendor: Optional[str] = None
    serial_number: Optional[str] = None
    os: Optional[str] = None
    os_version: Optional[str] = None
    hardware_model: Optional[str] = None
    ned_id: Optional[str] = None


class AssetClusterAsset(BaseModel):
    id: Optional[int] = None
    vendor: Optional[str] = None
    serial_number: Optional[str] = None
    os: Optional[str] = None
    os_version: Optional[str] = None
    hardware_model: Optional[str] = None
    ned_id: Optional[str] = None
    cluster_index: Optional[int] = None


class AssetCluster(BaseModel):
    id: Optional[int] = None
    type: Literal["asset_cluster"] = "asset_cluster"
    name: Optional[str] = None
    ned_id: Optional[str] = None
    assets: List[AssetClusterAsset] = []


class AssetRefType(str, Enum):
    asset = "asset"
    asset_cluster = "asset_cluster"


class NodeStatus(str, Enum):
    planned = "planned"
    init = "init"
    active = "active"
    decommissioned = "decommissioned"


class NodeBase(BaseModel):
    asset_ref_id: int
    asset_ref_type: Optional[AssetRefType] = None
    logical_node_id: int
    status: Optional[NodeStatus] = None

class Node(NodeBase):
    id: Optional[int] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


class NodeListItem(NodeBase):
    """Enriched list representation with denormalized fields from asset and logical_node."""
    id: Optional[int] = None
    # From logical_node
    hostname: Optional[str] = None
    site: Optional[str] = None
    # From asset
    vendor: Optional[str] = None
    os: Optional[str] = None
    ned_id: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


class NodeResponse(NodeBase):
    id: Optional[int] = None
    asset: Annotated[Union[Asset, AssetCluster], Field(discriminator="type")]
    logical_node: LogicalNodeResponse
    created_at: datetime
    updated_at: Optional[datetime] = None



"""Stateless rule engine and data types for config normalization.

Two independent passes:
    normalize_pass  — drops lines/blocks that are not intent-config
    mask_pass       — rewrites lines to redact secrets

The passes are isolated: normalize never touches content it keeps,
mask never drops lines. This makes the operations order-independent.
"""

from __future__ import annotations

import re
from dataclasses import dataclass, field


# ── Rule types ────────────────────────────────────────────────────────

@dataclass(frozen=True)
class LineRule:
    """Matches a full line → the line is dropped."""
    name: str
    pattern: re.Pattern


@dataclass(frozen=True)
class BlockRule:
    """Matches a block header → header + indented body are dropped."""
    name: str
    pattern: re.Pattern


@dataclass(frozen=True)
class RewriteRule:
    """Rewrites a line (e.g. redact secrets)."""
    name: str
    pattern: re.Pattern
    replacement: str


# ── Result types ──────────────────────────────────────────────────────

@dataclass
class OpStats:
    lines_in: int = 0
    lines_out: int = 0
    lines_dropped: int = 0
    blocks_dropped: int = 0
    lines_redacted: int = 0
    by_rule: dict[str, int] = field(default_factory=dict)

    def _bump(self, rule: str) -> None:
        self.by_rule[rule] = self.by_rule.get(rule, 0) + 1


@dataclass
class OpResult:
    config: str
    operation: str  # "normalize" or "mask"
    stats: OpStats


# ── Engine ────────────────────────────────────────────────────────────

class NormalizerEngine:
    """Stateless rule engine used by BaseNormalizer subclasses."""

    def normalize_pass(
        self,
        raw: str,
        *,
        line_rules: list[LineRule],
        block_rules: list[BlockRule],
        is_block_continuation,
        is_block_terminator,
    ) -> tuple[list[str], OpStats]:
        stats = OpStats()
        lines = raw.splitlines()
        stats.lines_in = len(lines)
        out: list[str] = []

        i = 0
        while i < len(lines):
            line = lines[i]

            block_hit = next(
                (r for r in block_rules if r.pattern.match(line)), None
            )
            if block_hit:
                stats.blocks_dropped += 1
                stats._bump(f"block:{block_hit.name}")
                stats.lines_dropped += 1
                i += 1
                while i < len(lines) and (
                    is_block_continuation(lines[i])
                    or is_block_terminator(lines[i])
                ):
                    stats.lines_dropped += 1
                    i += 1
                continue

            line_hit = next(
                (r for r in line_rules if r.pattern.match(line)), None
            )
            if line_hit:
                stats.lines_dropped += 1
                stats._bump(f"line:{line_hit.name}")
                i += 1
                continue

            out.append(line)
            i += 1

        return out, stats

    def mask_pass(
        self,
        raw: str,
        *,
        rewrite_rules: list[RewriteRule],
    ) -> tuple[list[str], OpStats]:
        stats = OpStats()
        lines = raw.splitlines()
        stats.lines_in = len(lines)
        out: list[str] = []

        for line in lines:
            rewritten = line
            for rule in rewrite_rules:
                new = rule.pattern.sub(rule.replacement, rewritten)
                if new != rewritten:
                    stats.lines_redacted += 1
                    stats._bump(f"redact:{rule.name}")
                    rewritten = new
                    break
            out.append(rewritten)

        stats.lines_out = len(out)
        return out, stats

    @staticmethod
    def collapse_and_trim(
        lines: list[str],
        stats: OpStats,
        comment_marker: str = "!",
    ) -> list[str]:
        collapsed: list[str] = []
        prev_marker = False
        for line in lines:
            is_marker = line.strip() == comment_marker
            if is_marker and prev_marker:
                stats.lines_dropped += 1
                stats._bump("post:collapse_comment")
                continue
            collapsed.append(line)
            prev_marker = is_marker

        while collapsed and collapsed[0].strip() in ("", comment_marker):
            collapsed.pop(0)
        while collapsed and collapsed[-1].strip() in ("", comment_marker):
            collapsed.pop()
        return collapsed



"""Models for ACE-X DevKit."""

# TODO: Move all models to this package
# TODO: Move all base classes to this package

from .external_value import ExternalValue
from .attribute_value import AttributeValue
from .node_response import NodeResponse, NodeListItem, LogicalNodeResponse
from .management_connection import ManagementConnection, ConnectionType

__all__ = [
    ExternalValue,
    AttributeValue,
    NodeResponse,
    NodeListItem,
    LogicalNodeResponse
    ]
